"""
Lattice that maps inputs to best-matching-units (BMU) of kernel.
             enables Kohonen-Layers aka Self-Organizing-Maps
                     also: non-parametric UMAP.
             is capable of high-dimensional Donut-topologies for cyclic data like angles.

Example 6-dimensional input mapped to 3-dimensional lattice:

        ltc = Lattice(output_shape=tf.constant([100, 100, 100]), isCyclic=[False, False, True])
        ltc.build(input_shape=[None, 6])
"""

import tensorflow as tf

class Lattice(tf.keras.layers.Layer):

    def __init__(self, output_shape, isCyclic):
        super(Lattice, self).__init__()
        self.grid_shape = output_shape
        self.grid = None
        self.mask_cyclic = None
        self.set_topology(isCyclic)
        pass

    def build(self, input_shape):
        self._init_meshgrid(self.grid_shape, indexing='ij')
        grid_flattened = tf.reshape(self.grid, shape=[1, -1, self.grid_shape.shape[0]])
        grid_flattened = tf.cast(grid_flattened, dtype=tf.float32)  # + tf.random.uniform(shape=grid_flattened.shape)/3.
        grid_flattened = (grid_flattened - tf.reduce_mean(grid_flattened, axis=1)) / tf.math.reduce_std(grid_flattened,
                                                                                                        axis=1)
        self.grid_flattened = tf.Variable(initial_value=grid_flattened, trainable=True)
        self.w = tf.Variable(initial_value=tf.random.normal(shape=[1,
                                                                   tf.math.reduce_prod(self.grid_shape),
                                                                   input_shape[1]
                                                                   ],
                                                            stddev=1),
                             trainable=False)
        super(Lattice, self).build(input_shape)
        pass

    def set_topology(self, isCyclic):
        isCyclic = tf.cast(isCyclic, dtype=tf.float32)
        # self.isCyclic = tf.reshape(tf.cast(isCyclic, dtype=tf.int32), shape=[1,-1])
        self.mask_cyclic = tf.reshape(isCyclic, shape=[1, -1])
        pass

    def _init_meshgrid(self, shape, indexing='ij'):
        ranges = tf.map_fn(tf.range, shape, fn_output_signature=tf.RaggedTensorSpec(shape=[None], dtype=tf.int32))
        grid = tf.meshgrid(*ranges, indexing=indexing)
        self.grid = tf.stack(grid, axis=-1)
        pass

    def _get_linear_grid_distance(self, x, y):
        distance = tf.abs(x - y)
        distance = distance * (1. - self.mask_cyclic)
        distance_linear = tf.reshape(distance, shape=[x.shape[0], -1, self.grid_shape.shape[0]])
        return distance_linear

    def _get_cyclic_grid_distance(self, x, y):
        maxi = tf.cast(self.grid_shape, dtype=tf.float32)
        centre = maxi / 2
        distance = tf.math.floormod(tf.abs(x - y) - centre, maxi) - centre
        distance = tf.abs(distance) * self.mask_cyclic
        distance_cyclic = tf.reshape(distance, shape=[x.shape[0], -1, self.grid_shape.shape[0]])
        return distance_cyclic

    def _get_index_distance(self, x, y, norm=2):
        distance = self._get_linear_grid_distance(x, y) + self._get_cyclic_grid_distance(x, y)
        distance = tf.norm(distance, ord=norm, axis=-1)
        return distance

    def _get_BMUs(self, x, hm_neighbors=1):
        # get best matching unit (BMU) of lattice
        k = hm_neighbors
        grid = self.grid_flattened
        x = tf.expand_dims(x, axis=1)

        activation_distance = tf.abs(x - self.w)
        activation_distance = tf.square(activation_distance)  # via l2-norm
        activation_distance = tf.reduce_sum(activation_distance, axis=-1)
        values_neg, indices = tf.math.top_k(-tf.abs(activation_distance), k)
        neighbors_distance = -1 * values_neg

        bmu = tf.gather(grid[0], indices[:, 0])

        return bmu, neighbors_distance

    def call(self, x):
        bmu, neighbors_distance = self._get_BMUs(x)
        return bmu

    def compute_output_shape(self, input_shape):
        return tf.constant(1).shape  # self.grid_shape


import tensorflow as tf
import keras_tuner as keras_tuner

class HyperModel(keras_tuner.HyperModel):

    """
    HyperModel-Factory for KerasTuner with ADAM and early-stopping
    """

    def __init__(self, models, in_shape, out_shape, loss=['mse'], metrics=['mse'], *args, **kwargs):
        """
        models:     List of functions that generate keras-models with hyperparameters
                    [model_1, model_2, ...] each with args: (hyperparameters, in_shape, out_shape)
        """
        self.model_choices = [model.__name__ for model in models]
        self.models = {model.__name__: model for model in models} 
        self.in_shape = in_shape
        self.out_shape = out_shape
        self.loss = loss
        self.metrics = metrics
        self.seed = None
        self.batch_size = None
        super(HyperModel, self).__init__(*args, **kwargs)

    def fit(self, model, *args, **kwargs):
        stop_early = tf.keras.callbacks.EarlyStopping(monitor='loss', patience=2, restore_best_weights=True)
        callbacks = kwargs['callbacks'] + [stop_early]
        epochs = kwargs['epochs']

        history = model.fit(
                            batch_size=self.batch_size,
                            callbacks=callbacks,
                            epochs=epochs,
                            *args,
                            )
        return history

    def compile_model(self, model, hp):
        self.seed = hp.Choice("seed", [27, 42, 123])
        tf.random.set_seed(self.seed)
        self.batch_size = hp.Choice("batch_size", [16, 128, 512, 1024])

        learning_rate = hp.Float("lr", min_value=1e-5, max_value=1e-1, sampling="log")
        beta_1 = hp.Float("beta_1", min_value=0.5, max_value=0.999, sampling="log")
        beta_2 = hp.Float("beta_2", min_value=0.7, max_value=0.9999, sampling="log")
        epsilon = hp.Choice("epsilon", [1e-7, 1e-5, 1e-3, .1, 1.])

        adam = tf.keras.optimizers.Adam(learning_rate=learning_rate,
                                        beta_1=beta_1,
                                        beta_2=beta_2,
                                        epsilon=epsilon,
                                        )

        model.compile(steps_per_execution=1,
                      optimizer=adam,
                      loss=self.loss,
                      metrics=self.metrics
                      )
        model._name = model.name + '_' + str(self.seed) + '_lr_' + str(learning_rate)
        return model

    def build(self, hp):
        model_type = hp.Choice("model_type", self.model_choices)
        model = self.models[model_type](hp, self.in_shape, self.out_shape)
        model = self.compile_model(model, hp)
        return model



import os
import numpy as np
from matplotlib import pyplot as plt
import IPython.display as ipd
import librosa


def print2txt(func, path2file):
  # Save func() output to .txt file
  with open(path2file, 'w') as file:
    # Redirect standard output to the file
    original_stdout = sys.stdout
    sys.stdout = file
    func()
    # Restore standard output
    sys.stdout = original_stdout
    pass


def print_plot_play(signal, sr=None):
    """
        - Prints & Plots information about an audio singal,
        - Creates player
    """
    if type(signal) == str:
      path2file = signal
      signal, sr = librosa.load(path2file, sr=sr)
      text = 'audiofile: ' + np.char.split(path2file, sep ='/').item(-1)[-1]
      print('%s Fs = %d, x.shape = %s, x.dtype = %s' % (text, sr, signal.shape, signal.dtype))

    fig, axs = plt.subplots(nrows=3, ncols=1, figsize=(12, 7))

    # Plot the signal
    axs[0].plot(signal, color='grey', linewidth=2)
    axs[0].set_xlabel('Time (samples)')
    axs[0].set_ylabel('Amplitude')
    axs[0].set_title('Time Domain')
    axs[0].grid(True)

    # Plot the phase spectrum
    axs[1].phase_spectrum(signal, Fs=sr, color='green', linewidth=2)
    axs[1].set_title('Phase Spectrum')
    axs[1].set_xscale('log')
    axs[1].grid(True, which='both', axis='x')

    # Plot the magnitude spectrum
    axs[2].magnitude_spectrum(signal, Fs=sr, scale='dB', color='blue', linewidth=2)
    axs[2].set_title('Magnitude Spectrum')
    axs[2].set_xscale('log')
    axs[2].set_ylim([-100, 10])
    axs[2].grid(True, which='both', axis='x')

    fig.tight_layout()
    plt.show()



    ipd.display(ipd.Audio(data=signal, rate=sr))

    return signal, sr



import tensorflow as tf

def connect(units=128):
    """
    Keras_Layer generator that:
                                1. concatenates multiple inputs
                                2. feeds outcome into new layer e.g. Dense
                                3. normalizes output batches

    Example:                    x4 = connect('dense')([x1, x2, x3])
    """

    concat = tf.keras.layers.Concatenate(axis=-1)
    normalize = tf.keras.layers.BatchNormalization()
    last_layer = tf.keras.layers.Dense(units, activation='relu')

    block = lambda x: last_layer(normalize(concat(x)))

    return block



from .helpers import *
from .paths import *
from .plot import *
from .generator import *
from .connect import *
from .hypermodel import *
from .training import *
from .text import *

from .chaos import *
from .custom_metrices import *
from .lattice import *
from .numeric import *
from .rounding_loss import *
from .spiral import *
from .voltera import *
from .sparse import *
from .triangle import *
from .polynom import *




import tensorflow as tf


class Spiral1D(tf.keras.layers.Layer):
    def __init__(self, units=32):
        super(Spiral1D, self).__init__()
        self.units = units
        self.koordinates = None
        self.size = None
        self.phi2 = None
        self.PI = tf.constant(3.1415926535897932)

    def build(self, input_shape):
        signal_length = tf.reduce_prod(input_shape[1:])
        signal_length_flt = tf.cast(signal_length, dtype=tf.float32)

        size = tf.sqrt(signal_length_flt / ((self.PI / 4.) * 0.70))
        size = tf.round(size / 2.)
        size = size * 2 + 1
        self.size = size

        r, phi = self.__create_circular_grid()

        phi2 = tf.round(r) * 2 * self.PI + phi
        self.phi2 = tf.cast(phi2, dtype=tf.float32)

        spiral = r / phi2
        spiral = tf.where(tf.math.is_nan(spiral), 0., spiral)

        koordinates = tf.reshape(phi2, shape=[-1])
        koordinates = tf.argsort(koordinates)

        koordinates = tf.expand_dims(koordinates, axis=1)
        koordinates = tf.cast(koordinates, dtype=tf.int64)

        self.koordinates = koordinates

    def call(self, inputs):
        signal = tf.reshape(inputs, shape=[-1])
        koordinates = self.koordinates[:signal.shape[0]]
        size = self.size

        size = tf.cast(size, dtype=tf.int32)
        size2 = tf.cast(size * size, dtype=tf.int32)

        spiral = tf.sparse.SparseTensor(indices=koordinates, values=signal, dense_shape=[size2])
        spiral = tf.sparse.reorder(spiral)
        spiral = tf.sparse.to_dense(spiral, default_value=None, validate_indices=True)
        spiral = tf.reshape(spiral, shape=[size, size])
        spiral = tf.stack([spiral, self.phi2], axis=-1)
        spiral = tf.expand_dims(spiral, axis=0)
        return spiral

    def __create_grid2d(self):
        size = tf.cast(self.size, dtype=tf.float32)
        rnge = tf.range(self.size)
        rnge = rnge - size / 2. + 0.5
        x1, x2 = tf.meshgrid(rnge, rnge)
        grid = tf.stack([x1, x2])
        return grid

    def __create_circular_grid(self):
        x1, x2 = self.__create_grid2d()
        r = tf.abs(x1 * x1 + x2 * x2)
        r = tf.sqrt(r)
        phi = x1 / r
        phi = tf.acos(phi)
        phi = tf.where(tf.math.is_nan(phi), 0., phi)
        phi = phi * tf.sign(x2)  # +PI
        mask1 = x2 == 0
        mask2 = x1 < 0
        is_pi = tf.math.logical_and(mask1, mask2)
        is_pi = tf.cast(is_pi, dtype=tf.float32) * self.PI
        phi = phi + is_pi
        return r, phi


class Spiral(tf.keras.layers.Layer):

    def __init__(self):
        super(Spiral, self).__init__()
        # self.units = units
        self.koordinates = None
        self.size = None
        self.phi2 = None
        self.PI = tf.math.acos(0.) * 2

    def build(self, input_shape):
        batch_size = tf.cast(input_shape[0], dtype=tf.float32)
        signal_length = tf.cast(input_shape[1], dtype=tf.float32)
        channel_dims = tf.cast(input_shape[-1], dtype=tf.float32)
        self.batch_size = batch_size
        self.signal_length = signal_length
        self.channel_dims = channel_dims

        size = tf.sqrt(signal_length / ((self.PI / 4.) * 0.70))
        size = tf.round(size / 2.)
        size = size * 2 + 1
        self.size = size
        size2 = tf.cast(size * size, dtype=tf.int32)
        self.size2 = size2

        indices_signal = self.__calc_signal_indices()
        indices = self.__combine_indices(indices_signal)
        indices = indices[:, :input_shape[1], :]
        indices = tf.reshape(indices, shape=[-1, 3])
        self.indices = indices

    def call(self, inputs):
        signal = tf.reshape(inputs, shape=[-1])
        spiral = tf.sparse.SparseTensor(indices=self.indices,
                                        values=signal,
                                        dense_shape=[inputs.shape[0],
                                                     self.size2,
                                                     inputs.shape[2]])
        spiral = tf.sparse.reorder(spiral)
        spiral = tf.sparse.to_dense(spiral, default_value=None, validate_indices=True)
        spiral = tf.reshape(spiral, shape=[inputs.shape[0], self.size, self.size, inputs.shape[2]])
        return spiral

    def get_embd(self):
        return tf.stack([self.phi2, self.r], axis=-1)

    def __create_grid2d(self):
        size = tf.cast(self.size, dtype=tf.float32)
        rnge = tf.range(self.size)
        rnge = rnge - size / 2. + 0.5
        x1, x2 = tf.meshgrid(rnge, rnge)
        grid = tf.stack([x1, x2])
        return grid

    def __create_circular_grid(self):
        x1, x2 = self.__create_grid2d()
        r = tf.abs(x1 * x1 + x2 * x2)
        r = tf.sqrt(r)
        phi = x1 / r
        phi = tf.acos(phi)
        phi = tf.where(tf.math.is_nan(phi), 0., phi)
        phi = phi * tf.sign(x2)  # +PI
        mask1 = x2 == 0
        mask2 = x1 < 0
        is_pi = tf.math.logical_and(mask1, mask2)
        is_pi = tf.cast(is_pi, dtype=tf.float32) * self.PI
        phi = phi + is_pi
        return r, phi

    def __calc_signal_indices(self):
        r, phi = self.__create_circular_grid()
        self.r = r

        phi2 = tf.round(r) * 2 * self.PI + phi
        self.phi2 = tf.cast(phi2, dtype=tf.float32)

        spiral = r / phi2
        spiral = tf.where(tf.math.is_nan(spiral), 0., spiral)

        indices_signal = tf.reshape(phi2, shape=[-1])
        indices_signal = tf.argsort(indices_signal)

        indices_signal = tf.expand_dims(indices_signal, axis=1)
        indices_signal = tf.cast(indices_signal, dtype=tf.int64)
        return indices_signal

    def __combine_indices(self, indices_signal):
        rnge_batch_size = tf.reshape(tf.range(self.batch_size, dtype=tf.int64),
                                     shape=[-1, 1, 1])

        indices_signal = tf.reshape(indices_signal,
                                    shape=[1, -1, 1])

        rnge_channel = tf.reshape(tf.range(self.channel_dims, dtype=tf.int64),
                                  shape=[1, 1, -1])

        ones = tf.ones([self.batch_size, indices_signal.shape[1], self.channel_dims], dtype=tf.int64)

        indices_batch = tf.einsum('bsc, boo->bsc', ones, rnge_batch_size)
        indices_signal = tf.einsum('bsc, oso->bsc', ones, indices_signal)
        indices_channel = tf.einsum('bsc, ooc->bsc', ones, rnge_channel)
        indices = tf.stack([indices_batch, indices_signal, indices_channel], axis=-1)
        return indices


import tensorflow as tf

def rapson_newton(fun, funprime, start, numinterations):
    """
    fun:          function that is monotone and zero at poit of interest
                           eg: lambda x:  x^3 - 5
    funprime:     first derivative of function.   eg: lambda x: 3x^2

    """
    estimate = start
    for n in range(numinterations):
        print(estimate)
        estimate = estimate - fun(estimate) / funprime(estimate)
    return estimate

@tf.function
def fun4n(fun, n, initial_state, external_force, const_param, dtype=tf.float32):
    states = tf.TensorArray(dtype, size=n, clear_after_read=False)
    state = initial_state

    for i in tf.range(n):
        state = fun(state, external_force[i], const_param)
        states = states.write(i, state)

    return states.stack()



import numpy as np
import tensorflow as tf

def standardize(x):
    x = (x-np.mean(x))/np.std(x)
    return x

def normalize(x):
    x = (x-np.min(x))/(np.max(x)-np.min(x))
    return x

def stretch(x, s):
    return x*[s, 1/s]

def flatten(x):
    hm_ins = x.shape[0]
    x = np.reshape(x, [hm_ins, -1])
    return x

def correlate(inputs, outputs, input_lag):
    lag = input_lag
    outputs = np.pad(outputs[:,lag:], pad_width=[[0,0], [0, lag]])
    coefs = np.corrcoef(inputs, outputs)
    return coefs[-1, :-1]

def group_unique(array, axis=0):
    a = array
    a = a[a[:, axis].argsort()]
    groups = np.split(a, np.unique(a[:, axis], return_index=True)[1][1:])
    return groups

def expspace(min, max, n=128, power=10):
  min = np.log(min)/np.log(power)
  max = np.log(max)/np.log(power)
  rnge = np.linspace(min, max, n)
  return np.power(power, rnge)

@tf.function
def rotate_deg(x, alpha_deg=45.):
    alpha = -alpha_deg/90.*tf.asin(1.)
    cos = tf.math.cos(alpha)
    sin = tf.math.sin(alpha)
    rot_matrix = [[cos, -sin], [sin, cos]]
    rotated_x = tf.linalg.matmul(x, rot_matrix)
    return rotated_x

def group_unique(array, axis=0):
    a = array
    a = a[a[:, axis].argsort()]
    groups = np.split(a, np.unique(a[:, axis], return_index=True)[1][1:])
    return groups


@tf.function
def span_polar_basis(x, y):
    phi = tf.math.atan2(y, x)
    e_r = tf.stack([
                    tf.math.cos(phi),
                    tf.math.sin(phi)
                   ])
    e_phi = tf.stack([
                      tf.math.sin(phi) * -1,
                      tf.math.cos(phi)
                     ])

    e_r = tf.cast(e_r, dtype=tf.float32)
    e_phi = tf.cast(e_phi, dtype=tf.float32)
    return e_r, e_phi


def create_meshgrid(shape = tf.constant([10, 10, 10]), indexing = 'ij'):
    ranges = tf.map_fn(tf.range, shape, fn_output_signature = tf.RaggedTensorSpec(shape=[None],dtype=tf.int32))
    grid = tf.meshgrid(*ranges, indexing = indexing)
    grid = tf.stack(grid, axis=-1)
    return grid



import tensorflow as tf


def huber_loss(y_true, y_pred, grad_max=1):
    """
    Robust Least-squares aka. Huber-penalty from
    Peter J. Huber. "Robust Estimation of a Location Parameter" - 1964

    inputs:
            y_true:     ground truth
            y_pred:     prediction
            grad_max:   maximum absolut gradient of loss

    outputs:
            mean of losses along last axis s.t.
            if |y_true - y_pred| <= threshold -> l2-penalty
            if |y_true - y_pred| > threshold -> l1-penalty
    """
    error = y_true - y_pred
    linear_loss = (2 * tf.abs(error) - grad_max) * grad_max
    linear_loss = tf.math.maximum(linear_loss, grad_max * grad_max)
    quadratic_loss = tf.square(error)
    loss = tf.math.minimum(linear_loss, quadratic_loss)
    return tf.reduce_mean(loss, axis=-1)


def deadzone_l1_loss(y_true, y_pred, error_min=0.01):
    """
    mean of l1 - losses with loss = 0 if |error| < error_min
    """
    error = tf.abs(y_true - y_pred)
    loss = tf.math.maximum(error - error_min, 0)
    return tf.reduce_mean(loss, axis=-1)


def complementary_distance(sample, batch, grad=100):
    """
    costfunction similar to interior point method
    data must encode boolean values via float
    eg: arg1 * arg2:  [1,0,1,1] * [[1,0,0,1]] -> low since arg2 does not complement 2. value of arg1
        arg2 * arg1:  [1,0,0,1] * [[1,0,1,1]] -> higher since arg1 complements 3. value of arg2

                                              -> Order matters!!!!!

    input:      sample of shape:[depth]
                batch of shape: [batchsize, depth]
    """
    batch = tf.cast(batch, dtype=tf.float32)
    grad = tf.cast(grad, dtype=tf.float32)
    order = 1./(tf.sqrt(1./grad))
    depth = batch.shape[-1]
    offset = tf.pow(order, 1/depth)
    offset = tf.exp(-offset)                               # TODO !!!!
    sample = tf.equal(sample, 0)
    cost = tf.boolean_mask(batch, sample, axis=1)
    cost = 1./(cost+(1./order))
    cost = cost/order
    cost = tf.reduce_sum(cost, axis=1) / cost.shape[-1]
    cost = tf.clip_by_value(cost, clip_value_min=0, clip_value_max=1)
    cost = tf.round(cost*grad)/grad
    return cost


import tensorflow as tf


def get_sampling_matrix(shape = tf.TensorShape([10, 100])):
    """
    returns sub-gaussian sampling-matrix
    """
    sampling_matrix = tf.random.uniform(shape)
    sampling_matrix = sampling_matrix / tf.expand_dims(tf.norm(sampling_matrix, axis=1), axis=1)
    return sampling_matrix

@tf.function
def lipschitz_grad_lasso(A):
    """
    calculates the smallest Lipschitz constant of grad(f(x)).
    With f(x) = l2(y - Ax) + l1(x); aka: LASSO.
    """
    At = tf.einsum('...ij -> ...ji', A)
    AtA = tf.einsum('...ji, ...ab -> ...jb', At, A)
    eig = tf.linalg.eig(AtA)[0]
    eig = tf.math.real(eig)
    eig_max = tf.math.reduce_max(eig, axis=-1)
    lipschitz = 2*eig_max
    return lipschitz


def soft_threshold(x, offset):
    '''
    prox for l1-norm
    '''
    thd = tf.abs(x) - offset
    gtr = tf.greater(thd, 0)
    gtr = tf.cast(gtr, dtype=tf.float32)
    thd = thd * gtr * tf.sign(x)
    return thd

def deadzone_l1_loss(y_true, y_pred, error_min=0.01):
    '''
    mean of l1 - losses with loss = 0 if |error| < error_min
    '''
    error = tf.abs(y_true - y_pred)
    loss = tf.math.maximum(error - error_min, 0)
    return tf.reduce_mean(loss, axis=-1)

def nesterov_momentum(t):
    '''
    momentum term defined by the increasing t sequence
    note: in order to increase performance of fista,
          it is best practise to reset the momentum
    '''
    t_new = (1+tf.sqrt(1+4*tf.square(t)))*0.5
    momentum = (t - 1)/t_new
    return momentum, t_new

@tf.function
def ista_prox_lasso(ground_truth, dictionary, estimation, rate=1):
    print('retrace')
    y = ground_truth
    A = dictionary
    x = estimation
    At = tf.einsum('...ij -> ...ji', A)
    Ax = tf.einsum('...ij, ...jo -> ...io', A, x)
    offset = 1 / lipschitz_grad_lasso(A)
    residual = Ax - y
    grad_l2 = 2 * tf.einsum('...ji, ...io -> ...jo', At, residual)
    arg = x - tf.einsum('j, jmn -> mn', offset, grad_l2)
    estimation = soft_threshold(arg, offset * rate)
    return estimation

@tf.function
def fista(y, A, x, t, rate=1):
    x_new = ista_prox_lasso(y, A, x, rate)
    m, t_new = nesterov_momentum(t)
    y = x_new + m * (x_new - x)
    return y, t_new

def lasso_l1(sigma=1e-2, cardinality=16):
    '''
    simga      :     Noise Level
    cardinaltiy:     of Dictionary
    '''
    sigma = tf.cast(sigma, dtype=tf.float32)
    cardinality = tf.cast(cardinality, dtype=tf.float32)
    rate = sigma * tf.sqrt(2. * tf.math.log(cardinality))
    rate = tf.cast(rate, dtype=tf.float32)
    l1_reg = tf.keras.regularizers.L1(float(rate))
    return l1_reg

def norm_l0(x, axis=None):
    l0 = tf.greater(tf.abs(x), 0)
    l0 = tf.cast(l0, dtype=tf.float32)
    l0 = tf.reduce_sum(l0, axis)
    return l0

def huber_loss(y_true, y_pred, grad_max=1):
    """
    Robust Least-squares aka. Huber-penalty from
    Peter J. Huber. "Robust Estimation of a Location Parameter" - 1964

    inputs:
            y_true:     ground truth
            y_pred:     prediction
            grad_max:   maximum absolut gradient of loss

    outputs:
            mean of losses along last axis s.t.
            if |y_true - y_pred| <= threshold -> l2-penalty
            if |y_true - y_pred| > threshold -> l1-penalty
    """
    error = y_true - y_pred
    linear_loss = (2 * tf.abs(error) - grad_max) * grad_max
    linear_loss = tf.math.maximum(linear_loss, grad_max * grad_max)
    quadratic_loss = tf.square(error)
    loss = tf.math.minimum(linear_loss, quadratic_loss)
    return tf.reduce_mean(loss, axis=-1)



import re
import nltk
import matplotlib.pyplot as plt
import tensorflow as tf
import numpy as np
import string
import os


def build_re_pattern(liste):
  """
  creates regular-expression pattern-string
  for a list of words.
  """
  pattern = ''.join([' '+w+r'\b|' for w in liste])[:-1]
  return pattern

def _load_data(file_name):
    dir_path = os.path.dirname(os.path.realpath(__file__))
    file_path = os.path.join(dir_path, 'data', file_name)
    return np.load(file_path)

_names_pattern = build_re_pattern(_load_data('english_names.npy'))
_opinionated_pattern = build_re_pattern(_load_data('opinionated.npy'))
_stopword_pattern = build_re_pattern(_load_data('stopwords.npy'))

def custom_standardization(input_data):
  """
  standadizes text strings
  eg: for text-vectorization via keras
  """
  stripped = tf.strings.regex_replace(input_data, _names_pattern, ' ')
  stripped = tf.strings.lower(stripped)
  stripped = tf.strings.regex_replace(stripped, _stopword_pattern, ' ')
  stripped = tf.strings.regex_replace(stripped, '<br />', ' ')
  stripped = tf.strings.regex_replace(stripped,
                                      '[%s]' % re.escape(string.punctuation),
  
                                      '')
  # clean newline and words that are split by newline
  # also kills digits
  stripped = tf.strings.regex_replace(stripped, r'\n|'+r'-\n|'+r'\t', '')
  # clean single letters
  stripped = tf.strings.regex_replace(stripped, '(^| ).(( ).)*( |$)', ' ')
  # replace numbers that are not part of a word
  # Note: needs to run 2x
  stripped = tf.strings.regex_replace(stripped, r'(\A|[. ,])\d+([. ,]|$)', '  ')
  stripped = tf.strings.regex_replace(stripped, r'(\A|[. ,])\d+([. ,]|$)', '  ')
  return stripped

def build_vocab(dataset_train, standardization, batchsize=128, max_tokens=130_000):
  """
  builds vocabulary from dataset with custom_standardization
  """
  vectorize = tf.keras.layers.TextVectorization(
    max_tokens=max_tokens,
    standardize=standardization,
    split='whitespace',
    output_mode='int')
  # create vocabulary from train_ds
  text_only_ds = dataset_train.map(lambda x, y: x).batch(batchsize)
  vectorize.adapt(text_only_ds)
  vocab = vectorize.get_vocabulary()
  return vocab

def stemm(vocab):
  """
  stemm via nltk.snowball
  """
  stemmed = [nltk.stem.SnowballStemmer("english").stem(word) for word in vocab]
  return stemmed

def plot_stemming_efficiency(stemmed, min_vocab_size = 3000):
  """
  helps with choosing the right min_vocab_size
  by plotting relative stemming efficiency.

  args:
        stemmed - list of stemmed vocabulary that is far to long 
        min_vocab_size - integer of current estimation that gets
                         marked in the plot
  """
  min_vocab_k = tf.cast(min_vocab_size/1000, dtype=tf.int32)
  stemming_efficiency = lambda i: ((i-len(set(stemmed[:i]))) / len(set(stemmed[:i])))
  stemming_efficiency = [stemming_efficiency(i) for i in range(1, len(stemmed), 1000)]
  plt.figure(figsize=(10, 3))
  plt.title('relative stemming efficiency:')
  plt.xlabel('size of vocabulary in 1000s')
  plt.ylabel('efficiency')
  plt.plot(stemming_efficiency)
  plt.plot(stemming_efficiency[:min_vocab_k])
  return min_vocab_size

def build_lookup(keys, values, min_keys=3000, verbose=False):
  """
  build lookup-dict from unique but non-bijective key-value pairs
  note: <keys> and <values>   should correspond to eachother and
                                     be ordered according to frequency

  args:
        keys           - list of keys that is far to long 
        values         - list of values that is far to long
        min_keys       - integer that describes the number of 
                         unique lookup values. Defaults to 3000,
                         wich is the average number of words an
                         educated not-native english-speaker knows

  returns: lookup-dict with at least <min_keys> key-value pairs
  
  """
  table = [list(z) for z in zip(keys, values) if z[1] in stemmed[:min_keys]]
  table = dict(table)
  if verbose:
    print(str(len(table.keys())) + ' different words will be mapped to ' + 
          str(len(set(table.values()))) + ' unique tokens')
  return table

def custom_split(text):
  """
  spitter that can be used within tf.keras.layers.TextVectorization

  """
  split = tf.strings.regex_replace(text, ' +', ' ')
  split = tf.strings.split(split, sep=' ')
  return split

def reduce_freq(look_up_table, min_keys=3000, verbose=False):

  """
  builds lookup-dictionaries from unique but non-bijective key-value pairs
                             that have reduced size

  note: <keys> and <values>  should correspond to eachother and
                                    be ordered according to frequency

        keys that have values that are not within
        the most frequent <min_keys> will return 0

  args:
        look_up_table[0]    - list of keys that shall be reduced
        look_up_table[1]    - list of values that shall be reduced
        min_keys            - integer that describes the number of
                              unique lookup values. Defaults to 3000,
                              wich is the average number of words an
                              educated not-native english-speaker knows

  returns: lookup-dictionaries with at least <min_keys> key-value pairs

           keys2values:
                              keys -> values for values wich have keys
                              that are most frequent (top <min_keys>)
           top_values2range:
                              values from keys2values -> integer
                              wich inreases approximatly
                              with decreasing key-frequency
           keys2int:
                              keys from keys2values -> int
                              according to top_values2range
           range2int:
                              like keys2values, but with integer-keys
                              and -values that correspond to frequency:
                              higher int -> lower frequency

  """

  keys, values = look_up_table

  zero = tf.constant(0, dtype=tf.int64)
  lookup_table = [list(z) for z in zip(keys, values) if z[1] in values[:min_keys]]
  keys2values = dict(lookup_table)
  unique_values = list(dict.fromkeys(keys2values.values()))
  top_values2range = dict(zip(unique_values,
                              1+tf.range(len(unique_values), dtype=tf.int64)))
  keys2int = {key: top_values2range.get(values[i])
                   for i, key in enumerate(keys)}
  keys2int = {key: value
                   for key, value in keys2int.items()
                   if value!=None}
  range2int = {i+1: keys2int.get(key) for i, key in enumerate(keys)
                                    if keys2int.get(key) != None}

  if verbose:
    print(str(len(lookup_table)) + ' different words will be mapped to ' +
          str(len(unique_values)) + ' unique tokens')

  return keys2values, top_values2range, keys2int, range2int


class LookUp(tf.keras.layers.Layer):
  
  """
  Class is depreciated and will be remvoved in future updates
  """
  
  def __init__(self, standardize, split, lookup):
    super(LookUp, self).__init__()
    self.standardize = standardize
    self.split = split
    keys = list(lookup.keys())
    values = list(lookup.values())
    values = tf.strings.to_hash_bucket_strong(values, len(values), [2, 42])
    lookup =  tf.lookup.KeyValueTensorInitializer(keys, values)
    self.lookup = tf.lookup.StaticVocabularyTable(lookup, num_oov_buckets = 1)
    self._name = 'lookup'
    
  def call(self, inputs):
    text = self.standardize(inputs)
    words = self.split(text)
    words = tf.ragged.map_flat_values(self.lookup.lookup, words)
    return words


# Dictionarys of standardize/split functions, class 'Tokenizer' is allowd to use:
allowed_standardize_funcs = dict({'custom_standardization': custom_standardization})
allowed_split_funcs = dict({'custom_split': custom_split})


class Tokenizer(tf.keras.layers.Layer):
  
  def __init__(self,  lookup, standardize='custom_standardization', split='custom_split', output_sequence_length=None):
    super(Tokenizer, self).__init__()
    self.standardize = standardize
    self.split = split
    self.lookup = lookup
    self.output_sequence_length = output_sequence_length
    
    self._standardize = allowed_standardize_funcs[standardize]
    self._split = allowed_split_funcs[split]

    self._keys = list(lookup.keys())
    self._values = list(lookup.values())
    _initializer =  tf.lookup.KeyValueTensorInitializer(self._keys, self._values)
    self._lookup = tf.lookup.StaticHashTable(_initializer, default_value=0)
    self._name = 'tokenizer'

  def pad_and_trunc(self, x):
    """
    input: ragged tensor with shape: [batchsize, None]
    output: tensor with shape: [batchsize, output_sequence_length]
    """
    x = x.to_tensor()
    shape = tf.shape(x)
    x = x[..., :self.output_sequence_length]
    shape = tf.shape(x)
    padded_shape = tf.concat((shape[:-1], [self.output_sequence_length]), 0)
    padding, _ = tf.required_space_to_batch_paddings(shape, padded_shape)
    x = tf.pad(x, padding)
    return x

  def call(self, inputs):
    text = self._standardize(inputs)
    words = self._split(text)
    words = tf.ragged.map_flat_values(self._lookup.lookup, words)
    if self.output_sequence_length != None:
      words = self.pad_and_trunc(words)
    return words

  def get_config(self):
    # Convert tensor values to a serializable format if necessary:
    lookup_serializable = {k: (v.numpy() if isinstance(v, tf.Tensor) else v) for k, v in self.lookup.items()}

    config = super(Tokenizer, self).get_config()
    config.update({
            'standardize': self.standardize,
            'split': self.split,
            'lookup': lookup_serializable,
            'output_sequence_length': self.output_sequence_length,
            })
    return config



def build_text_input(standardize,
                     split,
                     lookup_dict,
                     embedding_dims,
                     output_sequence_length,):
                       
  vocab_size = len(lookup_dict.values()) + 1
  input_layer = tf.keras.Input(shape=(1,), dtype=tf.string)
  lookup_layer = Tokenizer(standardize, split, lookup_dict, output_sequence_length)
  embedding_layer = tf.keras.layers.Embedding(vocab_size,
                                              embedding_dims,
                                              name='embedding')
  reshape = tf.keras.layers.Lambda(lambda x: tf.squeeze(x, axis=1))
  model = tf.keras.models.Sequential([input_layer,
                                      lookup_layer,
                                      embedding_layer,
                                      reshape,
                                      ])
  return model



import tensorflow as tf


class Polynom(tf.keras.layers.Layer):
    '''
    Polynom:      of order n is spanned for each channel:
                  x, x^2 ... , x^n

    Note:         This layer does not learn anything!
                  It only maps the input to a polynomial basis.

    input_shape:  [batch_size, sequence_length, channels]
    output_shape: [batch_size, sequence_length, channels, n]


    '''

    def __init__(self, units=32, eps=0.01, **kwargs):
        '''
        units:  order of Polynom
        eps:    max relative error for 99.98 % of Data,
                given that Data is ~ N(1, 0)
        '''
        super(Polynom, self).__init__(**kwargs)
        self.units = units
        self.eps = eps
        self.xpnt = None
        self.norm = None
        self.sig = None

    def build(self, input_shape):
        rnge = tf.range(self.units, dtype=tf.float32) + 1.
        self.xpnt = tf.reshape(rnge, shape=[1, 1, 1, -1])
        self.norm = 1. / self.integrate_xn(rnge)
        quantil = 3.719  # 0.9999 Quantil for N(0,1)
        self.sig = quantil / tf.sqrt(-2. * tf.math.log(1 - self.eps))

    @tf.function
    def factorial(self, n):
        fac = tf.exp(tf.math.lgamma(n + 1.))
        return fac

    @tf.function
    def integrate_xn(self, n):
        '''
        itegrates f(x, n, a) dx from 0 to inf
        with f(x, n, a) = x^n * exp(-0.5(x)^2)
        '''
        pwr = tf.pow(2., (n - 1) / 2)
        gma = self.factorial((n + 1) / 2 - 1)
        return pwr * gma

    @tf.function
    def window(self, inputs):
        """
        Gaußian ~ N(0, sig)
        """
        window = tf.exp(-0.5 * tf.square(inputs / self.sig))
        return window

    @tf.function
    def span(self, inputs):
        """
        spans a basis of x^n from n=[1, units] (without constant!)

        inputs_shape: [batch_size, sequence_length, channels]
        output_shape: [batch_size, sequence_length, channels, n]

        inputs:       is expected to be ~ N(0,1)

        output:       is normed s.t.: the expected amplitude
                      of its elements are equal.
                      is windowed via a Gaußian ~ N(0, sig[n]) in order to avoid extreme values
                      s.t.: the relative error to x^n is upperbounded to 'eps'
                      for 99.98 % of the inputs. Given that 'w' did converge!
        """
        inputs = tf.expand_dims(inputs, axis=-1)
        pwr = tf.pow(inputs, self.xpnt)
        window = self.window(inputs)
        inputs = pwr * window
        output = tf.einsum('bsck, k ->bsck', inputs, self.norm)
        return output

    @tf.function
    def call(self, inputs):
        output = self.span(inputs)
        return output

    def get_config(self):
        config = super().get_config()
        config.update({"units": self.units})
        return config


class Polynom_Dense(tf.keras.layers.Layer):
    '''
    Polynom:     is learned for each channel:
                 w1*x + w2*x^2 ... + wn*x^n + b

    input_shape: [batch_size, sequence_length, channels]
    '''

    def __init__(self, units=32, eps=0.01, **kwargs):
        '''
        units:  order of Polynom
        eps:    max relative error for 99.98 % of Data,
                given that Data is ~ N(1, 0)
        '''
        super(Polynom, self).__init__(**kwargs)
        self.units = units
        self.eps = eps
        self.xpnt = None
        self.norm = None
        self.sig = None

    def build(self, input_shape):
        rnge = tf.range(self.units, dtype=tf.float32) + 1.
        self.xpnt = tf.reshape(rnge, shape=[1, 1, 1, -1])
        self.norm = 1. / self.integrate_xn(rnge)
        quantil = 3.719  # 0.9999 Quantil for N(0,1)
        self.sig = quantil / tf.sqrt(-2. * tf.math.log(1 - self.eps))

        self.w = self.add_weight(
            shape=(input_shape[-1], self.units),
            initializer="random_normal",
            trainable=True,
        )

        self.b = self.add_weight(
            shape=(1, 1, input_shape[-1]),
            initializer="random_normal",
            trainable=True,
        )

    @tf.function
    def factorial(self, n):
        fac = tf.exp(tf.math.lgamma(n + 1.))
        return fac

    @tf.function
    def integrate_xn(self, n):
        '''
        itegrates f(x, n, a) dx from 0 to inf
        with f(x, n, a) = x^n * exp(-0.5(x)^2)
        '''
        pwr = tf.pow(2., (n - 1) / 2)
        gma = self.factorial((n + 1) / 2 - 1)
        return pwr * gma

    @tf.function
    def window(self, inputs):
        """
        Gaußian ~ N(0, sig)
        """
        window = tf.exp(-0.5 * tf.square(inputs / self.sig))
        return window

    @tf.function
    def span(self, inputs):
        """
        spans a basis of x^n from n=[1, units] (without constant!)

        inputs_shape: [batch_size, sequence_length, channels]
        output_shape: [batch_size, sequence_length, channels, n]

        inputs:       is expected to be ~ N(0,1)

        output:       is normed s.t.: the expected amplitude
                      of its elements are equal.
                      is windowed via a Gaußian ~ N(0, sig[n]) in order to avoid extreme values
                      s.t.: the relative error to x^n is upperbounded to 'eps'
                      for 99.98 % of the inputs. Given that 'w' did converge!
        """
        inputs = tf.expand_dims(inputs, axis=-1)
        pwr = tf.pow(inputs, self.xpnt)
        window = self.window(inputs)
        inputs = pwr * window
        output = tf.einsum('bsck, k ->bsck', inputs, self.norm)
        return output

    @tf.function
    def call(self, inputs):
        inputs = self.span(inputs)
        output = tf.einsum('bsck, ck->bsck', inputs, self.w)
        output = tf.reduce_sum(output, axis=-1) + self.b
        return output

    def get_config(self):
        config = super().get_config()
        config.update({"units": self.units})
        return config



import os
import random

import numpy as np
import pandas as pd

class ListedPaths:
    def __init__(self, directory=''):
        self.directory = directory
        self.list = []
        self.position = 0
        self.__gather()

    def __gather(self):

        for subdir, dirs, files in os.walk(self.directory):
            for file in files:
                self.list.append(os.path.join(subdir, file))
        self.list.sort()
        pass

    def next2(self):
        next2 = [self.list[self.position], self.list[self.position+1]]
        self.position += 1
        return next2
    
    def get_endings(self):
        endings = [path.split('.')[-1] for path in self.list]
        endings = list(dict.fromkeys(endings))        
        return endings
    
    def get_rdm(self, k=1):
        rdm_paths = random.sample(self.list, k)
        return rdm_paths

    def __len__(self):
        return len(self.list)
    
    def __add__(self, arg):
        self.list += arg.list
        self.list = list(dict.fromkeys(self.list))
        summ = ListedPaths()
        summ.list = self.list
        return summ
    
    def __mul__(self, arg):
        self.list = [path for path in self.list if path.split('.')[-1] == arg]
        prodct = ListedPaths()
        prodct.list = self.list
        return prodct
    
    def __call__(self):
        return self.list    
    
    
def strings2onehot(strings):
    df = pd.Categorical(strings, ordered=False)
    onehot = pd.get_dummies(df).to_numpy()
    return df.to_numpy(), onehot     
    
def paths2labels(paths, seperators=['/', '.'], indices=[-2, 0]):
    labels = [np.char.split(lst, sep =seperators[1])[indices[0]][indices[1]] for lst in np.char.split(paths, sep =seperators[0])]
    return labels

def paths2label_dicts(paths, seperators=['/', '.'], indices=[-2, 0]):
    labels = paths2labels(paths, seperators, indices)
    labels = list(dict.fromkeys(labels))
    keys, labels = strings2onehot(labels)
    labels_dict = dict(zip(keys, labels))
    return labels_dict

def map_via_dict(key_list, label_dict):
    return [label_dict[key] for key in key_list]
  
  
 



import tensorflow as tf

class Voltera(tf.keras.layers.Layer):
    """
    input_shape:  [batch_size, sequence_length, channels]
    output_shape: [batch_size, sequence_length]

    Voltera:      1. For each channel a polynomial-basis is spanned:
                     1, x, x^2 ... , x^n
                  2. Every single x^i basis (channels * (n+1))
                     are combined via an outerproduct
                  3. Every product gets weighted
                  4. All weighted products are summed

    """

    def __init__(self, units=16, outputs=1, eps=0.001, kernel_regularizer=None, **kwargs):
        '''
        units:  defines order of Polynom: (units 16 -> order=15 -> 16^channels weights)
        eps:    max relative error for 99.98 % of Data,
                given that Data is ~ N(1, 0)
        '''
        super(Voltera, self).__init__(**kwargs)
        self.units = units
        self.order = units - 1
        self.outputs = outputs
        self.eps = eps
        self.xpnt = None
        self.norm = None
        self.sig = None
        self.equation = None
        self.nr_weights = None
        self.kernel_regularizer = tf.keras.regularizers.get(kernel_regularizer)

    def build(self, input_shape):

        self._input_shape = input_shape
        rnge = tf.range(self.order, dtype=tf.float32) + 1
        self.xpnt = tf.reshape(rnge, shape=[1, 1, 1, -1])
        self.norm = 1. / self.integrate_xn(rnge)
        # 0.9999 Quantil for N(0,1)
        quantil = 3.719
        self.sig = quantil / tf.sqrt(-2. * tf.math.log(1 - self.eps))
        self.equation = self.einsum_equation(input_shape[-1])

        nr_weights = tf.pow(self.units, input_shape[-1])
        self.nr_weights = tf.cast(nr_weights, dtype=tf.int32)
        self.w = self.add_weight(shape=[self.nr_weights, self.outputs],
                                 initializer=tf.keras.initializers.RandomUniform(minval=-1e-6, maxval=1e-6),
                                 regularizer=self.kernel_regularizer,
                                 trainable=True
                                 )

    @tf.function
    def factorial(self, n):
        fac = tf.exp(tf.math.lgamma(n + 1.))
        return fac

    @tf.function
    def integrate_xn(self, n):
        '''
        itegrates f(x, n, a) dx from 0 to inf
        with f(x, n, a) = x^n * exp(-0.5(x)^2)
        '''
        pwr = tf.pow(2., (n - 1) / 2)
        gma = self.factorial((n + 1) / 2 - 1)
        return pwr * gma

    def einsum_equation(self, order):
        '''
        generates string for einsum notation for outerproduct of last dimensions
        eg.: '...x, ...y, ...z->...xyz'
        '''
        xyz = 'xyzuvwijklmnopqrstfghabcde' + 'ABSCDEFGHIJKLMNOPQRSTUVWXYZ'
        equation = ''
        for i in range(order):
            equation = equation + '...' + xyz[i]
            if i == order - 1:
                break
            equation = equation + ', '
        equation = equation + '->...'
        for i in range(order):
            equation = equation + xyz[i]
        return equation

    @tf.function
    def outer_product(self, inputs):
        '''
        calculates batch-wise outer-product of last dimension
        input_shape: [batch_size, sequence, channels, order]

        '''
        xyz = tf.unstack(inputs, axis=-2)
        output = tf.einsum(self.equation, *xyz)
        output = tf.reshape(output, shape=[-1, self._input_shape[1], self.nr_weights])
        return output

    @tf.function
    def window(self, inputs):
        '''
        Gaußian ~ N(0, sig)
        '''
        window = tf.exp(-0.5 * tf.square(inputs / self.sig))
        return window

    @tf.function
    def span_polynoms(self, inputs):
        '''
        spans a basis of x^n from n=[0, units]

        inputs_shape: [batch_size, sequence_length, channels]
        output_shape: [batch_size, sequence_length, channels, n]

        inputs:       is expected to be ~ N(0,1)

        output:       is normed s.t.: the expected amplitude
                      of its elements are equal.
                      is windowed via a Gaußian ~ N(0, sig[n]) in order to avoid extreme values
                      s.t.: the relative error to x^n is upperbounded to 'eps'
                      for 99.98 % of the inputs. Given that 'w' did converge!
        '''
        inputs = tf.expand_dims(inputs, axis=-1)
        pwr = tf.pow(inputs, self.xpnt)
        window = self.window(inputs)
        inputs = pwr * window
        output = tf.einsum('bsck, k ->bsck', inputs, self.norm)
        # Adding a constant to the Polynom for every channel
        ones = 1. + 0 * output[:, :, :, 0]
        ones = tf.expand_dims(ones, axis=-1)
        output = tf.concat([ones, output], axis=-1)
        return output

    @tf.function
    def span(self, inputs):
        inputs = self.span_polynoms(inputs)
        outputs = self.outer_product(inputs)
        return outputs

    @tf.function
    def call(self, inputs):
        inputs = self.span(inputs)
        inputs = tf.einsum('bsk, ko ->bso', inputs, self.w)
        return inputs

    def get_config(self):
        config = super().get_config()
        config.update({"units": self.units})
        return config


import tensorflow as tf

@tf.function
def saddle(x, radius=1., mode=1., mass=0.5, sensitivity=1.):
    """
    saddle-node bifurcation
    inputs: x            state
            radius(+/-)  defines the bistable positions for modes(+/-)
            mode         switches the stable radius according to sign(mode)
            mass         defines the acceleration to/from modes
                         for high mass: the absolute speed between modes
                         is almost constant!!!
            sensitivity  scales mode

    output: x_dot        speed between modes
                               is normed s.t.: max(speed) = 1!
    """
    print("Traced")
    mass = tf.abs(mass) + 1.
    speed = tf.pow(tf.abs(radius), mass)
    x = tf.pow(tf.abs(x), mass)
    radius = tf.pow(radius, mass)
    mode = tf.tanh(mode * 3. * tf.abs(sensitivity))

    x_dot = (radius - x) * mode / speed
    return x_dot


@tf.function
def lorenz(state, force, parameters):
    """
    https://en.wikipedia.org/wiki/Lorenz_system
    """
    print('Traced')
    s = parameters[0]
    r = parameters[1]
    b = parameters[2]

    step = parameters[3]

    x = state[0]
    y = state[1]
    z = state[2]

    x_dot = s * (y - x)
    y_dot = r * x - y - x * z
    z_dot = x * y - b * z

    x = step * x_dot + x
    y = step * y_dot + y
    z = step * z_dot + z
    return tf.stack([x, y, z])


import tensorflow as tf

class GeneratorDataset():
    """
    
    Class that - enables tf.data.Dataset from generator that utilize random variables that are unique
               - is a work-around and might get obsolete with future versions of TensorFlow
    
    __call__:  returns a tf.data.Dataset with non-deterministic order of generated data.
    
    """
    @tf.autograph.experimental.do_not_convert()
    def __init__(self, generator, batch_size=1024, epochs=16, autotune=tf.data.AUTOTUNE, **kwargs):
        
        """
        
        'generator' is function like:
        
        def generator(batch_size, **kwargs):
        
            ...generate some random batch...
            
            return x, y
            
        'epochs' = maximum possible number of epochs 
        
        """
        
        self.generator = generator
        self.batch_size = batch_size
        self.epochs = epochs
        self.autotune = autotune
        
        dummy_generator = lambda: tf.range(self.epochs*self.batch_size, dtype=tf.int32)
        dataset = tf.data.Dataset.from_generator(dummy_generator, tf.int32)
        dataset = dataset.map(lambda dummy: self.__wrap(dummy, self.generator, **kwargs), num_parallel_calls=self.autotune)
        dataset = dataset.batch(batch_size, drop_remainder=True, num_parallel_calls=self.autotune, deterministic=None,)
        self.dataset = dataset.prefetch(self.autotune)
        pass
    
    def __wrap(self, dummy, core, **kwargs):
        x, y = core(batch_size=1, **kwargs)
        return (x[0], y[0])
    
    def __call__(self):
        return self.dataset
    



import tensorflow as tf

class Rounding_Loss(tf.keras.layers.Layer):

    """
    Regulizer that forces results to a linear grid
    """

    def __init__(self, resulution=1., rate=1e-2, power=1.):
        super(Rounding_Loss, self).__init__()
        self.rate = rate
        self.resulution = resulution
        self.power = power
        self.PI = tf.acos(0.) * 2

    def calc_round_loss(self, inputs):

        x = tf.tensordot(self.resulution, inputs, axes=0)
        loss = tf.cos(x * 2 * self.PI)
        loss = (loss + 1) / 2
        loss = 1 - tf.pow(loss, self.power)
        loss = tf.reduce_sum(loss, axis=-1)
        loss = loss / tf.cast(loss.shape[-1], dtype=tf.float32)
        loss = tf.reduce_mean(loss)

        return loss

    def call(self, inputs):
        self.add_loss(self.rate * self.calc_round_loss(inputs))
        return inputs

    def get_config(self):
        return {'resulution': self.resulution, 'rate': self.rate, 'power': self.power}


import tensorflow as tf


class Triangle(tf.keras.layers.Layer):

    def __init__(self, units=32, prime=3.):
        super(Triangle, self).__init__()
        self.units = units
        self.prime = prime

    def build(self, input_shape):
        w_init = tf.keras.initializers.HeUniform()
        self.w = tf.Variable(initial_value=w_init(shape=(input_shape[-1], self.units),
                                                  dtype='float32'), trainable=True)
        self.b = tf.Variable(initial_value=tf.zeros([self.units]), trainable=True)

    def call(self, inputs):
        inputs = inputs
        mapped = tf.matmul(inputs, self.w) + self.b
        result = tf.math.floormod(mapped, 4) - 2
        result = tf.abs(result) - 1
        return result



import tensorflow as tf
import matplotlib.pyplot as plt


def get_learning_rate(step=0, max_lr=0.1, max_step=1_000_000., resets=7, t1=1.5, t2=3., p1=2., p2=1.5, verbose=False):
    """
    Creates a learning-rate that diminishes over 'step' in a saw-like fashion.
                            for callbacks.
    Note: Use 'verbose=True' to get accustomed to the parameterization of its shape!
    """
    if verbose:
        step = tf.range(max_step)

    step = tf.cast(step + 1, dtype=tf.float32) / max_step
    saw = tf.pow(step, p1)
    saw = tf.math.floormod(-saw, 1 / (resets + 1))
    saw = saw / (1. / (resets + 1.))  # saw gets normalized
    saw = tf.exp(saw * t2)
    saw = saw / tf.exp(t2)  # saw gets normalized (again!)
    saw = saw * tf.pow(1 - step, p2)
    lr = max_lr * (tf.exp(-step * t1) + saw) / 2

    if verbose:
        plt.plot(lr)

    return lr
